use std::env;
use std::cmp::{min, max};
use std::time::{Duration,Instant};

use statrs::distribution::{LogNormal, Continuous, ContinuousCDF, Uniform, Normal};

//use statrs::statistics::Distribution;
use rand::distributions::Distribution;
use rand::prelude::*;

struct Model{
    lb: f64,
    ub: f64,
    divisor: f64,
    dist: LogNormal
}

impl Model {
    fn new(mean: f64,
	   sd: f64,
	   lb: f64,
	   ub: f64) -> Model {
	assert!(ub > lb);

	let mut m = Model{lb: lb,
			  ub: ub,
			  divisor: 0.0,
			  dist: LogNormal::new(mean, sd).unwrap()};
	m.divisor = m.dist.cdf(ub) - m.dist.cdf(lb);
	return m;
    }

    fn pdf(&self, x:f64) -> f64 {
	if x < self.lb || x > self.ub {
	    0.0
	} else {
	    self.dist.pdf(x) / self.divisor
	}
    }

    fn pdf_debug(&self, x:f64) -> f64 {
	if x < self.lb || x > self.ub {
	    println!("out of bound {} < {} < {}", self.lb,x,self.ub);
	}
	self.pdf(x)
    }
}

fn sd(v: &Vec<f64>) -> f64 {
    let mean = v.iter().sum::<f64>() / (v.len() as f64);
    return v.iter().map(|x:&f64| (*x - mean).powi(2)).sum::<f64>().sqrt();
}

fn storm_hyeto_likelihood(models: &Vec<Option<Model>>, org_par: &Vec<f64>, MAX: f64, sum_push:f64, debug:bool) -> f64 {
    let mut r: f64 = 0.0;
    let mut probs:Vec<f64> = vec![];

    for dur in 1..=min(models.len(), org_par.len() - 1) {
	if let Some(model) = &models[dur-1] {

	    if debug {
		println!("model {} div: {}",dur, model.divisor);
	    }

	    for j in 1..=(org_par.len() - dur + 1) {
		let sum: f64 = org_par[j-1..=(j+dur-1)-1].iter().sum();
		let p = if debug {model.pdf_debug(sum)} else {model.pdf(sum)};
		probs.push(p);
		if debug {
		    println!("    p: {}",p);
		}
		if p > 0.0 {
		    // Avoid the non negative logs.
		    r += (p+30.0*0.1).log(10.0);
		} else { // p == 0
		    // Avoid infinte value of log
		    r += -9999999999999.9;
		}
	    }
	}
    }

    let r_prob = r;
    let sum_org_par:f64 = org_par.iter().sum::<f64>();

    // FIXME Sum push is not correct
    let distance = (MAX-sum_org_par).abs().powi(2);
    // let f = if sum_push < 100.0 { 100.0 - sum_push } else { 0.0 };
    let f = 1.0;
    r = r * (f+distance);

    // let sdc = sd(&probs);
    // r = r * (1.0+sdc);

    // let mut vmin = 10000.0;
    // let mut vmax = 0.0;
    // for x in probs {
    // 	if x < vmin {
    // 	    vmin = x
    // 	} else if x > vmax {
    // 	    vmax = x
    // 	}
    // }
    // r = r + (vmin -vmax).abs();

    // let entropy:f64 = org_par.iter().map(|x:&f64| if *x == 0.0 {0.0} else {x*x.log(10.0)}).sum();
    // r = r / (1.0 + entropy);
    let entropy = 0; // for when you want to disable

    // Standard Deviation is useless
    // Imagine you have these diffs: 4 1 1 or 2 2 2, they all amoun tot the same std...
    // let mean = sum_org_par / (org_par.len() as f64);
    // let sd = org_par.iter().map(|x:&f64| (*x - mean).powi(2)).sum::<f64>();
    // r = r * (40.0+sd/mean); // 30 looks good but fort average, 40 makes extremes

    if debug {
	println!("r_prob={} r={} distance: {} entropy: {}", r_prob, r, distance, entropy);
    }
    return r;
}

fn likelihood(models: &Vec<Option<Model>>, org_par: &Vec<f64>, MAX: f64, sum_push: f64) -> f64 {
    storm_hyeto_likelihood(&models, org_par, MAX, sum_push, false)
}

fn prior(models: &Vec<Option<Model>>, org_par: &Vec<f64>) -> f64 {
    let model = models[0].as_ref().unwrap();
    org_par.iter().map(|x| model.pdf(*x).log(10.0)).sum::<f64>()
}

fn proposalfunction(models: &Vec<Option<Model>>, org_par: &Vec<f64>, MAX:f64) -> Option<Vec<f64>> {
    let mut rng = rand::thread_rng();
    let mut shrink:f64 = 1.0;
    let mut nb_iter = 0;
    for j in 1..3 {
	loop {
	    let p:Vec<f64> = org_par.iter().map(
		|x|
		if *x <= 0.2 { // between 0.1 and 0.2 --> between 0.1 and 0.4
		    Uniform::new(0.1, 0.1+0.3*shrink).unwrap().sample(&mut rng)
		} else {
		    Normal::new(*x, 2.0*shrink).unwrap().sample(&mut rng)
		}
	    ).collect();
	    if p.iter().find(|&&x| x < 0.1).is_none() // All mm are >= NOISE
	    //&& p.iter().sum::<f64>() <= MAX // The hyetogram is possible FIXME not correct ! si on peut aller en dessous de MAX, on doit pouvoir aller au dessus...
	    {
		return Some(p);
	    }
	    // 0.999**1000  = 0.36
	    // 0.999**10000 = 0
	    shrink = shrink * 0.999;
	}
    }

    return None;
}

fn posterior(models: &Vec<Option<Model>>, org_par: &Vec<f64>, MAX:f64, sum_push: f64) -> f64 {
    likelihood(models, org_par, MAX, sum_push) + prior(models, org_par)
}


fn run_metropolis_MCMC(models: &Vec<Option<Model>>, start_value: &Vec<f64>, iterations:usize, MAX: f64, sum_push: f64) -> Vec<Vec<f64>> {
    let mut rng = rand::thread_rng();
    let mut chain:Vec<Vec<f64>> = vec![];
    chain.push(start_value.to_vec());
    let runif = Uniform::new(0.0, 1.0).unwrap();
    for i in 1..=iterations {
	let previous = chain.last().unwrap();
        if let Some(proposal) = proposalfunction(models, previous, MAX) {
            let probab = (posterior(models, &proposal, MAX, sum_push) -
			  posterior(models, previous, MAX, sum_push)).exp();
	    if runif.sample(&mut rng) < probab {
		chain.push(proposal)
	    } else {
		chain.push( chain.last().unwrap().to_vec() );
	    }
	} else {
	    return chain;
	}
    }
    return chain;
}

fn MCMC_loop(models: &Vec<Option<Model>>, long_storm_duration: usize, MAX:f64, debug:bool) -> Vec<f64> {
    let mut rng = rand::thread_rng();
    let mut begin_time = Instant::now();

    let mut sum_push:f64 = 1.0;
    loop {
	// let s = (0..long_storm_duration).map(
	//     |x|
	//     Uniform::new(0.1, MAX).unwrap().sample(&mut rng)).collect();
	// let start_value:Vec<f64> = proposalfunction(
	//     models,
	//     &s,
	//     MAX);

	// Start from average and then move rain from one hour to another.
	let mut s = vec![MAX/(long_storm_duration as f64); long_storm_duration];
	for i in 1..(10*long_storm_duration) {
	    let ndx: usize = Uniform::new(0.0, long_storm_duration as f64).unwrap().sample(&mut rng).trunc() as usize;
	    let ndx2: usize = Uniform::new(0.0, long_storm_duration as f64).unwrap().sample(&mut rng).trunc() as usize;

	    let delta:f64 = 0.1;
	    if s[ndx] > 0.1 + delta && s[ndx2] < MAX - delta {
		s[ndx] -= delta;
		s[ndx2] += delta;
	    }
	}

	//return vec![MAX/(long_storm_duration as f64); long_storm_duration];

	let start_value = s;

	const BURNIN:usize = 50000;

	let chain = run_metropolis_MCMC(models, &start_value, BURNIN+long_storm_duration*1000, MAX, sum_push);

	// Read the chain backward until we find a hystogram
	// with the requested total rainfall (MAX mm)
	let candidate = chain.iter().rev().find(|&x| {
	    let s = x.iter().sum::<f64>();
	    s >= MAX-0.1 && s <= MAX+0.1
	});


	if let Some(c) = candidate {
	    return c.to_vec();
	} else {
	    if begin_time.elapsed() > Duration::from_secs(1*60) { // 15 minutes
		// Too long, bailing output
		return vec![MAX/(long_storm_duration as f64); long_storm_duration]
	    } else {
		sum_push = sum_push*2.0;
		if debug {
		    println!("No luck");
		    println!("{}", (chain.last().unwrap()).iter().fold(String::new(), |acc, &num| acc + &num.to_string() + ", "));
		    println!("LL={}",
			     storm_hyeto_likelihood(models, chain.last().unwrap(), MAX, sum_push, true))
		}
	    }

	}
    }
}

fn main() {

    // dlnorm(x=4,meanlog=5, sdlog=2) == 0.009747557
    // println!("{}", LogNormal::new(5.0, 2.0).unwrap().pdf(4.0)); // 0.009747556600619484


    let mut models: Vec<Option<Model>> =vec![];
    // 1.0,2.0,0.0,23.0,2.0,2.0,0.0,23.0,3.0,2.0,0.0,23.0,4.0,3.0,0.0,23.0,5.0,3.0,0.0,23.0,6.0,3.0,0.0,23.0,
    let args: Vec<String> = env::args().collect();

    let long_storm_duration = args[1].parse::<usize>().unwrap();
    let max_mm = args[2].parse::<f64>().unwrap();

    let mut i = args[3].split(",");
    while let Some(t) = i.next() {
	let t1 = t.parse::<f64>().unwrap();
	let t2 = i.next().unwrap().parse::<f64>().unwrap();
	let t3 = i.next().unwrap().parse::<f64>().unwrap();
	let t4 = i.next().unwrap().parse::<f64>().unwrap();
	models.push(Some(Model::new(t1,t2,t3,t4)))
    }

    // let v:Vec<f64> = vec![1.1, 2.1, 3.0, 2.0];
    // storm_hyeto_likelihood(&models, &v, 8.0, 1.0, true);

    let h = MCMC_loop(&models, long_storm_duration, max_mm, false);
    println!("{}",
	     h.iter().fold(String::new(), |acc, &num| acc + &num.to_string() + ","));
}
