options(error = function() {
    traceback(3)
    #if (!interactive()) q()
})

library(argparser)
library(ggplot2)
library(patchwork) # sum ggplots
library(copula)
source("raingen.R")

mix_pdf <- function(x,c,w,s1,s2, loc) {
    x <- x - loc
    # Not defined in x==0 if c < 1 (because c-1 < 0 => 1/x !)!
    c*w*(x**(c-1))/(s2**c) * exp(-(x/s2)**c) + (1-w)/s1*exp(-x/s1)
}
mix_cdf <- function(x,c,w,s1,s2,loc) {
    x <- x - loc
    1 - (1-w)*exp(-x/s1) - w*exp(-(x/s2)**c)
}
dmix <- mix_pdf
pmix <- mix_cdf
ll <- function(par, data) {
    ## c <- par[4]
    ## w <- par[1]
    ## s1 <- par[2]
    ## s2 <- par[3]
    c <- par[1]
    w <- par[2]
    s1 <- par[3]
    s2 <- par[4]
    loc <- par[5]
    t <- mix_pdf(data,c,w,s1,s2,loc)
    t[t < 1e-100] = 1e-100
    r <- sum(log(t))
    r
}
fit_mix <- function(data) {
    o <- optim(par=c(2/3,0.5,1,1,0),
               fn=ll,
               method="L-BFGS-B",
               lower=c(0.001, 0.0001,    1,    1, 0),
               upper=c(   10,      1, 1000, 1000, min(data)*0.999),
               control = list(fnscale=-1),
               data=data) # fnscale for maximisation
    return(o$par)
}



#  From:
# https://www.itl.nist.gov/div898/handbook/eda/section3/eda366b.htm
# The formula for the percent point function of the gamma distribution
# does not exist in a simple closed form. It is computed numerically.

dmix2 <- function(x,w,s1,s2, r1, r2, loc) {
    #x <- unlist(x, use.names = FALSE)
    x <- x - loc
    r <- w*dgamma(x,shape=s1,rate=r1)+(1-w)*dgamma(x,shape=s2,rate=r2)
    r
}
pmix2 <- function(x, w,s1,s2, r1, r2, loc) {
    x <- x - loc
    r <- w*pgamma(x,shape=s1,rate=r1)+(1-w)*pgamma(x,shape=s2,rate=r2)
    r
}
ll2 <- function(par, data) {
    t <- dmix2(data,par[1],par[2],par[3],par[4],par[5],par[6])
    t[t < 1e-100] = 1e-100
    r <- sum(log(t))
    r
}
fit2_mix <- function(data) {
    z <- optim(par=c(0.25,0.1,1,1,1,0),
               fn=ll2,
               method="L-BFGS-B",
               lower=c(0.0001,      0,      0,    0,    0,  0),
               upper=c(     1,     10,      10,   10,   10, min(data)*0.099999),
               control = list(fnscale=-1),
               data=data)
    print(paste("Best:", z$value))
    return(z$par)
}
make_qmix2 <- function(data, params) {
    min_x <- min(data$Duration)
    max_x <- max(data$Duration)
    rng <- seq(min_x, max_x*2, length.out=10000)
    cdf <- pmix2(rng, params[1], params[2], params[3], params[4], params[5], params[6])
    f <- function(p, ...) {
        approxfun(
            cdf, # probabilities
            rng, # actual values
            yleft=min_x, yright=max_x)(p)
    }
    return(f)
}


############################################################
# ALL STORMS MODELS
############################################################

print("--------------------------------------------------------------------")
library(stringr)
ap <- arg_parser("Raingen")
ap <- add_argument(ap, "--nb-sims", help="number of simulations", default=10)
ap <- add_argument(ap, "--nb-years", help="number of years", default=20)
ap <- add_argument(ap, "--shapes", help="shapes: optim_h,peaky,mcmc_h,flat,triangle", default="mcmc_h")
ap <- add_argument(ap, "--storms-only", help="just storms, no hyetograms", default=TRUE)
ap <- add_argument(ap, "--station-name", help="number of simulation", default="BASTOGNE")
ap <- add_argument(ap, "--dry-run", help="dry run", default=FALSE)
argv <- parse_args(ap)

## orig_storms <- read.table("all.csv",sep=",",header=TRUE)
## orig_storms$Date <- ymd_hms(orig_storms$Date)
## orig_tseries <- read.table("time_series_all.csv",sep=",",header=TRUE)
## orig_tseries$Date <- ymd_hms(orig_tseries$Date)
## print("Read files")

#args <- c("BASTOGNE")
station_name <- argv$station_name

file_name <- paste("time_series_",station_name,".csv", sep="")
print(paste("Loading original data",file_name))
full_ts <- read.table(file_name,sep=",",header=TRUE)
full_ts$Date <- ymd_hms(full_ts$Date)
data <- simple_storm_separation(full_ts)

print(paste(nrow(data),"storms spanning +/-",
            round(difftime( max(data$Date), min(data$Date), units="weeks") / 52.18,2), "years, SID=",
            string_to_station_id[[station_name]]))

full_storms_data <- extract_storms_hyetograms(c(string_to_station_id[[station_name]]))
peak_from_averages_model <- compute_model_peak_from_averages2(full_storms_data$all_long_storms)
cut_storm_models <- plot_subdurations( subdurations(full_storms_data$all_long_storms),900)
cut_storm_models_log_norm <- plot_subdurations_log_norm( subdurations(full_storms_data$all_long_storms),900)

# We need orignal TS because we need to track noise
noise_model <- make_noise_model(full_ts,
                                noise_level=NOISE_LEVEL,
                                min_precipitations_per_storm=1)


set.seed(3)

simulate_storms <- function(data, nb_storms) {
    # Distribution for the precipitation marginal
    fmpar <- fit_mix(data$mm)
    c <- fmpar[1]
    w <- fmpar[2]
    s1 <- fmpar[3]
    s2 <- fmpar[4]
    loc <- fmpar[5]

    # Building all the usual distribution functions for the
    # inteisity PDF

    # The quantile function (inverse of CDF) is a bit tough though.
    d <- seq(min(data$mm), max(data$mm)*1.2, length.out=10000)
    # qmix_helper is defined here and only here. So if the call
    # to qmix() works, then it means it resolves to qmix_helper here
    # eventhough qmix_helper is in this scope but qmix is not...
    qmix_helper <- approxfun(mix_cdf(d, c, w, s1, s2, loc), d, yleft=min(d), yright=max(d))
    assign("qmix", function(x, c, w, s1, s2, loc) {
        qmix_helper(x)
    }, envir = .GlobalEnv)


    # Distribution for the duration marginal

    par_dur <- fit2_mix(data$Duration)
    #qmix2 <- make_qmix2(data, par_dur)
    assign("qmix2", make_qmix2(data, par_dur), envir = .GlobalEnv)

    # Fitting the copula now
    gumbel.cop <- gumbelCopula(param=2, dim = 2)
    fit.ml <- fitCopula(gumbel.cop, pobs(data[,c("Duration","mm")]), method = "ml")

    # For mvdc, I need margins fucntion: 'dnorm', 'pnorm', and 'qnorm' are all available.
    # I have d and p; I need q...

    multivariate_dist <- mvdc(
        copula=fit.ml@copula,
        margins=c("mix2","mix"),
        paramMargins = list(
            list(par_dur[1],par_dur[2],par_dur[3],par_dur[4],par_dur[5],par_dur[6]),
            list(c,w,s1,s2,loc)))

    # *1.1 to account for rounding issues
    samples <- as.data.frame(rMvdc( round(nb_storms*1.1), multivariate_dist))
    names(samples) <- c("Duration", "mm")

    samples$Duration <- round(samples$Duration)
    samples$mm <- round(samples$mm,1)
    samples <- samples[(samples$Duration >= 1) & (samples$mm >= 0.1),]
    samples <- samples[1:nb_storms,]
    print(paste("simulate_storms() : Generated", nrow(samples), "storms,", nb_storms, "were requested."))
    print(paste("simulate_storms() : debug", qmix2(0.5)))
    return(samples)
}

##################################################################################
### THE SIMULATION
##################################################################################


nb_simulations <- argv$nb_sim
nb_years_to_simulate <- argv$nb_years
start_date <- ymd("2100/1/1")
storm_shapes <- if(argv$storms_only) {
                    # No need to compute more than that; shape is only for heytograms
                    # But don't mix storm only simulation with heytogram simulation
                    c("storm")
                } else {
                    #storm_shapes <- c("flat","triangle","peaky","optim_h","mcmc_h")
                    #storm_shapes <- c("flat")
                    #storm_shapes <- c("optim_h", "peaky","mcmc_h","flat","triangle")
                    storm_shapes <- lapply(strsplit(argv$shapes,","), str_trim)[[1]]
                    #storm_shapes <- c("mcmc_h")
                    #c("peaky","mcmc_h","flat","triangle")
                }

set.seed(34)


#Counting number of storms in each month-of-season/year
mean_storms_per_winter <- mean(table(year(data[month(data$Date) %in% WINTER_MONTHS,]$Date)))
std_storms_per_winter  <- sd(  table(year(data[month(data$Date) %in% WINTER_MONTHS,]$Date)))
mean_storms_per_summer <- mean(table(year(data[month(data$Date) %in% SUMMER_MONTHS,]$Date)))
std_storms_per_summer  <- sd(  table(year(data[month(data$Date) %in% SUMMER_MONTHS,]$Date)))

print(paste("Nb of storms per summer, mean",round(mean_storms_per_summer,1),"std",round(std_storms_per_summer,1)))
print(paste("Nb of storms per winter, mean",round(mean_storms_per_winter,1),"std",round(std_storms_per_winter,1)))

for( storm_shape in storm_shapes) {

    if (argv$dry_run == FALSE) {
        file.remove(Sys.glob(paste("simstorm_",station_name,"_",storm_shape,"_*.csv",sep="")))
        file.remove(Sys.glob(paste("hyeto_",station_name,"_",storm_shape,"_*.csv",sep="")))
    } else {
        print("Dry Run: not deleting anything!")
    }

for(nb_simulation in seq(nb_simulations)) {
    now <- Sys.time()
    print("-----------------------------------------------------------------")
    print(paste(station_name, "Simulation number", nb_simulation, "out of", nb_simulations, "Shape:", storm_shape))
    print("-----------------------------------------------------------------")
    print("Generating seasons of more storms than needed, will be picked later,in right numba'")
    # Compute winter and summer base data
    seasons <- list(
        list(nb_storms_mean=mean_storms_per_winter,
             nb_storms_std=std_storms_per_winter,
             months=WINTER_MONTHS,
             days=182, # number of days in the season
             storms=simulate_storms(
                 data[month(data$Date) %in% WINTER_MONTHS,],
                 round(mean_storms_per_winter*nb_years_to_simulate*1.5)),
             storm_ndx=1),
        list(nb_storms_mean=mean_storms_per_summer,
             nb_storms_std=std_storms_per_summer,
             months=SUMMER_MONTHS,
             days=183,
             storms=simulate_storms(
                 data[month(data$Date) %in% SUMMER_MONTHS,],
                 round(mean_storms_per_summer*nb_years_to_simulate*1.5)),
             storm_ndx=1)
    )
    year_df <- data.frame()
    for (nb_year in seq(nb_years_to_simulate)) {
        #print(paste("Year",nb_year, "Storm NDX: winter=", seasons[[1]]$storm_ndx, "summer=", seasons[[2]]$storm_ndx))
        for (season_ndx in seq(2)) {
            # Get season's paramaeters and precomputed box storms
            season <- seasons[[season_ndx]]
            #nb_storms <- rpois(n=1,lambda=season$nb_storms_)
            nb_storms <- round(rnorm(n=1, mean=season$nb_storms_mean, sd=season$nb_storms_std))
            #print(paste("nb_storms this year and season =",nb_storms))
            samples <- season$storms[season$storm_ndx:(season$storm_ndx+nb_storms-1),]
            seasons[[season_ndx]]$storm_ndx <- seasons[[season_ndx]]$storm_ndx + nb_storms
            names(samples) <- c("Duration", "mm")
            # Storms are spread on the season. I put one storm per day.
            # This works only because the number of storms is quite smaller
            # than the number of days in a season (about one third)
            samples$Date <- sample(
                do.call("c", # https://stackoverflow.com/questions/15659783/why-does-unlist-kill-dates-in-r
                        sapply(season$months,
                               function(m) dmy(paste(seq(MONTHS_DAYS[m]),"-",
                                                     m,"-",
                                                     nb_year+year(start_date)-1,sep="")))),
                size=nb_storms,
                replace=FALSE)
            ## samples$Date <- sample(
            ##     # FIXME This is not correct! The first 3 months of the first year
            ##     # will be empty !
            ##     dmy(paste("1-",season$months[1],"-",nb_year+year(start_date)-1,sep="")) + seq(season$days),
            ##     size=nb_storms,
            ##     replace=FALSE)
            samples$StationId <- rep(BASTOGNE, nrow(samples))
            year_df <- rbind(year_df, samples)
        } # season loop
    } # year loop
    if (max(year_df$Date) > start_date + years(nb_years_to_simulate)) {
        original_size <- nrow(year_df)
        year_df <- year_df[year_df$Date + hours(year_df$Duration) < start_date + years(nb_years_to_simulate),]
        print(paste("Storms were generated outside of the requested period. Trimmed from ", original_size,
                    "down to", nrow(year_df)))
    }
    if (argv$dry_run == FALSE) {
        print(paste("Writing", nrow(year_df), "storms"))
        for(sndx in seq(2)) {
            in_season <- table(year(
                year_df[month(year_df$Date) %in%
                        if(sndx == 1) {WINTER_MONTHS} else {SUMMER_MONTHS},]$Date))
            print(paste("Storm number for Season", if(sndx == 1) {"winter"} else {"summer"},
                        "mean", round(mean(in_season),1),
                        "std",round(sd(in_season),1)))
        }
        fname  <- paste("simstorm_",station_name,"_",sprintf("%05d", nb_simulation),".csv", sep="")
        write.csv(year_df,
                  fname,
                  row.names=FALSE)
        print(paste("Wrote",fname))
    } else {
        print("Dry Run: not writing storm file !")
    }
    if(!argv$storms_only) {
        print(paste("Generating hyetogram for", nrow(year_df), "storms from",min(year_df$Date), "to", max(year_df$Date)))
        hyeto <- if (storm_shape %in% c("optim_h","mcmc_h")) {
                     full_sim_with_cuts(cut_storm_models_log_norm, year_df, nb_years_to_simulate, start_date, storm_shape)
                 } else {
                     storms_to_hyetogram_with_shape(year_df, nb_years_to_simulate, start_date, peak_from_averages_model, shape=storm_shape)
                 }
        mm_before_noise <- sum(hyeto$mm)
        hyeto <- add_noise(hyeto, noise_model)
        print(paste("Reinjected noise. Total precipitation before:",round(mm_before_noise),"mm, after", round(sum(hyeto$mm)),"mm. Ratio",round(sum(hyeto$mm)/mm_before_noise - 1,4),"% more"))
        print("Rounding to 0.1mm")
        hyeto$mm <- round(hyeto$mm,1)
        #plot(hyeto$mm,type="l")
        #print(hyeto)
        #readline()
        fname <- paste("hyeto_",station_name,"_",storm_shape,"_",sprintf("%05d", nb_simulation),".csv", sep="")
        if (argv$dry_run == FALSE) {
            write.csv(hyeto,
                      fname,
                      row.names=FALSE)
        } else {
            print("Dry Run: not writing anything!")
        }
        print(paste("Wrote",fname))
        s <- simple_storm_separation(hyeto, NOISE_LEVEL, min_mm=1, debug=TRUE)
        print(paste("The generated hyeto has", nrow(s), "storms"))
        for(sndx in seq(2)) {
            print(seasons[[sndx]]$nb_storms_mean)
            print(seasons[[sndx]]$nb_storms_std)
            in_season <- table(year(
                s[month(s$Date) %in%  if(sndx == 1) {WINTER_MONTHS} else {SUMMER_MONTHS},]$Date))
            print(paste("Storm number for Season", if(sndx == 1) {"winter"} else {"summer"},
                        "mean", round(mean(in_season),1), "versus", round(seasons[[sndx]]$nb_storms_mean,1),
                        "std",round(sd(in_season),1),"versus", round(seasons[[sndx]]$nb_storms_std,1)))
        }
    } else {
        print(paste("Storms only", argv$storms_only))
    }

    print(paste("Writing csv's, took", Sys.time() - now))
} # nb sim
} # shapes
