import re
import time
import os.path
from datetime import datetime, timedelta, date
from math import log, sqrt
from scipy.interpolate import interp1d
from statsmodels.distributions.empirical_distribution import ECDF

import matplotlib.pyplot as plt
from matplotlib import cm
import dateparser
import numpy as np
import pandas as pd
from tqdm import tqdm

NAN_THRESHOLD=100
MONTH_NAMES=["Jan","Fev","Mar","Avr","Mai","Jun","Juil","Aou","Sept","Oct","Nov","Dec"]

BALMORAL = 65370015
BASTOGNE = 99480015
BATTICE = 56490015
BIERSET = 72070015
BUTGENBACH = 69670015
CINEY = 81570015
EREZEE = 70480015
FLAMIERGES = 61470015
GEMMENICH = 52840015
JALHAY = 65500015
LANAYE = 57570015
LANDENNE=71680015
LOUVEIGNE=66570015
MARCHE = 70870015
MODAVE = 72280015
NASSOGNE = 85380015
ORTHO = 61280015
OUFFET = 70160015
RACHAMPS = 60480015
REULAND = 99140015
ROBERTVILLE = 69580015
ROCHEFORT = 85180015
SAINT_HUBERT = 61680015
SANKT_VITH = 99220015
TAILLES = 68580015
TERNELL=64970015 # aka NEU HATTLICH
VIELSALM = 68480015
WAREMME = 55780015
WAVRE=10430015

# All stations available on website

ANSEREMME=80630015
WASMUEL=35340015
ARLON=95960015
ATHUS=97810015
AUBANGE=97940015
AWANS=55960015
BALMORAL=65370015
BASTOGNE=99480015
BATTICE=56490015
BEAURAING=84680015
BERTRIX=96320015
BIERSET=72070015
BLAREGNIES=36490015
BOUILLON=94690015
BOUSSU=91370015
BOUSVAL=10810015
BUTGENBACH=69670015
CHATELET=73690015
CHIEVRES=29930015
CINEY=81570015
COMINES=38850015
COO=67120015
COO=67180015
CROIX=86870015
CRUPET=81380015
CUL=92880015
DAVERDISSE=83480015
DERGNEAU=24590015
ENGHIEN=28920015
EREZEE=70480015
FLAMIERGES=61470015
FLORENNES=81890015
FRATIN=95740015
GEDINNE=86770015
GEMMENICH=52840015
GERPINNES=76780015
HELECINE=15840015
JALHAY=65500015
KAIN=32770015
LANAYE=57570015
LANDENNE=71680015
LIBIN=83880015
LIGNY=76290015
LILLOIS=23480015
LOUVAIN=10570015
LOUVEIGNE=66570015
MARCHE=70870015
MEAN=70180015
MEIX=95880015
MODAVE=72280015
MOMIGNIES=75770015
MONCEAU=73950015
MORNIMONT=73350015
MOUSCRON=34760015
NAMOUSSART=96980015
NASSOGNE=85380015
ORTHO=61280015
ORVAL=98160015
OUFFET=70160015
PERUWELZ=37170015
PERWEZ=15790015
PETIGNY=92150015
PLATE=78650015
RACHAMPS=60480015
ROBERTVILLE=69580015
ROCHEFORT=85180015
ROISIN=36280015
ROUVEROY=36470015
SAINT_GERARD=81280015
SAINT_HUBERT=61680015
SART_TILMAN=70030015
SELANGE=97970015
SENEFFE=23880015
SENZEILLES=78880015
SIVRY=79670015
SOIGNIES=19970015
SOLRE=74850015
SOMME=70680015
SPA=65380015
STEFFESHAUSEN=99150015
STRAIMONT=96520015
SUGNY=96170015
TAILLES=68580015
TORGNY=97430015
TRIVIERES=35720015
TUBIZE=19540015
UCCLE=18980015
VEDRIN=72960015
VIELSALM=68480015
VRESSE=94360015
WAREMME=55780015


REGULAR_DATA_SET=[
    BALMORAL,BASTOGNE,BATTICE,BIERSET,BUTGENBACH,CINEY,
    EREZEE,FLAMIERGES,GEMMENICH,JALHAY,LANAYE,LANDENNE,
    LOUVEIGNE,MARCHE,MODAVE,NASSOGNE,ORTHO,OUFFET,RACHAMPS,
    REULAND,ROBERTVILLE,ROCHEFORT,SAINT_HUBERT,SANKT_VITH,
    SART_TILMAN,TAILLES,TERNELL,VIELSALM,WAREMME,WAVRE]

class Station:
    def __init__(self, ndx, row):
        self.sid = row.StationId
        self.x = row.X / 10000.0
        self.y = row.Y / 10000.0
        self.name = row.Name
        self.full_name = None
        self.ndx = ndx # 0 based

    def __lt__(self, other):
        return self.name < other.name

    def distance(self, s):
        return sqrt((self.x-s.x)**2 + (self.y-s.y)**2)

    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name

def ecdf(data):
    print("Deprecated! USes statsmodel equivalent")
    # Compute ECDF
    x = np.sort(data)
    n = x.size
    y = np.arange(1, n+1) / n
    return x,y

def invert_cdf(sample):
    sample_ecdf = ECDF(sample)
    slope_changes = sorted(set(sample))
    sample_edf_values_at_slope_changes = [ sample_ecdf(x) for x in slope_changes]
    inverted_edf = interp1d(sample_edf_values_at_slope_changes, slope_changes,
                            bounds_error = False,
                            fill_value=( np.min(sample_edf_values_at_slope_changes),
                                         np.max(sample_edf_values_at_slope_changes)) )
    return inverted_edf

def moving_average(x, w):
    return np.convolve(x, np.ones(w), 'valid') / w

def dmy(dt):
    return dt.strftime("%d/%m/%Y")

def human_date(s):
    # Specify the timezone to avoid PytzUsageWarning: The
    # localize method is no longer necessary, as this time
    # zone supports the fold attribute (PEP 495).
    return dateparser.parse(
        s,
        settings={'PREFER_DAY_OF_MONTH': 'first',
                  'TIMEZONE': 'Europe/Brussels'})

def hperiod(start_month, start_year, duration):
    # 2000-12 + 1 month => 2000-12-01, 2001-01-01

    b = np.datetime64(f'{start_year}-{start_month:02d}-01')
    if duration + start_month <= 12:
        return b, np.datetime64(f'{start_year}-{start_month+duration:02d}-01')
    else:
        return b, np.datetime64(f'{start_year+1}-{start_month+duration-12:02d}-01')



def myqqplot(a,b,line45=True,ax=None, **kwargs):
    if ax == None:
        axe = plt.gca()
    else:
        axe = ax

    q = np.linspace(0,1,max(len(a),len(b)))
    axe.scatter(np.quantile(a,q),
                np.quantile(b,q), **kwargs)

    m = min(np.max(a), np.max(b))
    if line45:
        axe.plot([0,m],[0,m], c="red")

    # axe.xlim(0, m)
    # axe.ylim(0, m)




def load_station_data(fpath):

    # Read station data
    STATION_DATA = pd.read_csv(fpath, header=0)
    # print(STATION_DATA)
    #print(STATION_DATA.dtypes)
    stations = dict()
    s_ndx = 0
    for ndx, row in STATION_DATA.iterrows():
        sid = int(row.StationId)
        assert sid  not in stations
        stations[sid] = Station(s_ndx, row)
        s_ndx += 1

    z = set()
    for s in stations.values():
        z.add(s.ndx)
    assert len(z) == max(z) + 1, f"{len(z)} <> {max(z)}"

    # Make python code
    # a = [(row.Name, row.StationId) for ndx, row in STATION_DATA.iterrows()]
    # for n, sid in sorted(a):
    #     print(f"{n} = {sid}")
    return stations


def load_data(DATAPATH, extended=False):

    # Panda code to read from CSV
    # start_time = time.perf_counter()
    # data = pd.read_csv(series_file, parse_dates=[3])
    # print(f"Loaded {series_file} in {time.perf_counter() - start_time:.1f} seconds")

    # FIXME Had to separate into three numpy arrays to
    # improve speed of loading... Sounds strange.

    if not extended:
        pfx = f"{DATAPATH}/np_"
    else:
        pfx = f"{DATAPATH}/np_all_"

    data = pd.DataFrame({
        "StationId": np.load(f"{pfx}station_id.npy"),
        "mm" : np.load(f"{pfx}mm.npy"),
        "Date" : np.load(f"{pfx}date.npy")})

    return data


def broad_select_bang(data, stations):
    # FIXME Side effects!

    nb_stations_orig = len(data["StationId"].unique())

    stations_to_drop = set()
    for sid in stations:
        ds = data[data["StationId"] == sid]
        nb_nans = np.count_nonzero(np.isnan(ds['mm']))
        if nb_nans > NAN_THRESHOLD:
            stations_to_drop.add(sid)

    print(f"Dropping {', '.join([stations[sid].name for sid in stations_to_drop])} because they have too many NaN's")
    for sid in stations_to_drop:
        stations.pop(sid)

    # Reindex the stations
    for s_ndx,s in enumerate(stations.values()):
        s.ndx = s_ndx

    data.drop( data[data["StationId"].isin(stations_to_drop)].index, inplace=True)

    data.fillna({"mm":0}, inplace=True)

    print(f"{len(data)} records left in {len(stations)} stations out of {nb_stations_orig} after drops and NaN's clear. Max mm of precipitation per hour is {np.max(data['mm'])}.")


def ins_read_excel(DATAPATH, loccode):
    #Lecture des caractéristiques d'une commune depuis le fichier Excel associé au code INS

    path = os.path.join(DATAPATH, "irm", f"{loccode}.xlsx")
    returnlevel = pd.read_excel(path,0,skiprows=range(7),nrows=19,usecols="A:M")
    standarddev = pd.read_excel(path,1,skiprows=range(7),nrows=19,usecols="A:M")
    confintlow  = pd.read_excel(path,2,skiprows=range(7),nrows=19,usecols="A:M")
    confintup   = pd.read_excel(path,3,skiprows=range(7),nrows=19,usecols="A:M")
    montana     = pd.read_excel(path,4,skiprows=range(11),nrows=12,usecols="A:G")

    # FIXME Convert the index into actual durations
    returnlevel = returnlevel.set_index("Duur / Durée")

    return returnlevel,standarddev,confintlow,confintup,montana



def filter_by(data: pd.DataFrame, sid, begin="1/1/2000", end="31/12/2022", col="mm"):
    # Will return either a numpy column vector or a pandas DataFrame.

    #print(f"Filter by sid={sid}, begin={begin} end={end}")

    if begin is None:
        begin = date(1970,1,1)
    elif type(begin) == str:
        begin = human_date(begin)
    elif type(begin) == date:
        begin = datetime(begin.year, begin.month, begin.day)

    assert(type(begin) == datetime)

    if end is None:
        end = date(2030,12,31)
    elif type(end) == str:
        end = human_date(end)
    elif type(end) == date:
        end = datetime(end.year, end.month, end.day)

    assert(type(end) == datetime)

    # FIXME Double check hours minutes seconds are not causing
    # any off by something issues.

    f = True
    if sid is not None:
        f = data["StationId"] == sid
    if begin is not None:
        f = f & (data['Date'] >= begin)
    if end is not None:
        f = f & (data['Date'] < end)

    filtered_data = data[f]
    if col in  ("mm", "Date", "StationId"):
        r = np.array(filtered_data[col])
        # Make it a column vector
        r = r[...,np.newaxis]
        ret = r

    elif col == "all":
        # Return a pandas DataFrame
        ret = filtered_data

    print(f"Filter by sid={sid}, begin={dmy(begin)} end={dmy(end)}. Result: from {dmy(np.min(filtered_data['Date']))} to {dmy(np.max(filtered_data['Date']))}")
    return ret


def zero_runs(a):
    # From https://stackoverflow.com/questions/24885092/finding-the-consecutive-zeros-in-a-numpy-array

    # Create an array that is 1 where a is 0, and pad each end with an extra 0.
    iszero = np.concatenate(([0], np.equal(a, 0).view(np.int8), [0]))
    absdiff = np.abs(np.diff(iszero))
    # Runs start and end where absdiff is 1.
    ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
    return ranges

def non_zero_runs(a):
    # From https://stackoverflow.com/questions/24885092/finding-the-consecutive-zeros-in-a-numpy-array
    # Create an array that is 1 where a is 0, and pad each end with an extra 0.
    iszero = np.concatenate(([0], np.not_equal(a, 0).view(np.int8), [0]))
    absdiff = np.abs(np.diff(iszero))
    # Zero-runs start and end where absdiff is 1.
    ranges = np.where(absdiff == 1)[0].reshape(-1, 2)
    return ranges



def compute_shypre_storms(base_data, daily_minimum=4, storm_minimum=20, noise_level=0.1,
                          plot=False, to_file=False):
    df_all_periods = None

    for sid in base_data['StationId'].unique():
        data = base_data[base_data['StationId'] == sid]['mm'].to_numpy()

        # Clear the noise
        data[data < noise_level] = 0

        #print(f"{len(data)} records")
        start_year = base_data['Date'].dt.year.min()
        start_time = base_data['Date'].min().to_numpy()

        # From hourly to daily precipitations

        daily_mm = []
        for i in range(0,len(data),24):
            daily_mm.append(np.sum(data[i:i+24]))
        daily_mm = np.array(daily_mm)

        # Filter out daily amounts which are too small
        daily_mm[daily_mm < daily_minimum] = 0

        # Count rain episodes -------------------------------------------------
        # (all days >= 4mmm and one is >= 20mm)

        episodes_per_year = dict()

        rain_episodes = []
        nzeros = non_zero_runs(daily_mm)
        for i in range(nzeros.shape[0]):
            h_storm_begin = nzeros[i,0]
            h_storm_end = nzeros[i,1] # storm's last day is this index *minus* 1
            if np.max(daily_mm[h_storm_begin:h_storm_end]) >= storm_minimum :
                rain_episodes.append( (h_storm_begin*24,h_storm_end*24) )

                y = start_year + h_storm_begin // 365 # FIXME Approx !
                episodes_per_year[y] = episodes_per_year.get(y,0) + 1


        # At this point, I have all the valid rain episodes.

        nb_periods_per_episode = []
        nb_showers_per_period = []
        shower_durations = []
        shower_mean_intensities = []
        shower_max_to_positions = []
        shower_max_to_means = []
        dry_zones = []
        all_periods = None

        # Episodes are now split in periods. ----------------------------------
        # Episode = successive days with rain >= 4mm.
        # BUT inside a day, it is quite possible to have *hours*
        # without rain.
        # So we split them into hourly period. In a period
        # it rains on every hours.

        for episode_begin, episode_end in rain_episodes:
            episode = data[episode_begin:episode_end]

            # In a period, it rains on every single hour.
            # So we split to keep only non zeros runs.
            periods = non_zero_runs(episode)
            assert np.sum((periods[:,1] - periods[:,0]) == 0) == 0

            # period times is measured according to beginning
            # of episode => we offset that to get real beginning dates
            # back.
            p2 = periods + episode_begin
            if all_periods is None:
                all_periods = p2
            else:
                all_periods = np.vstack( [all_periods, p2])

            nb_periods_in_episode = len(periods) # this counts the number of rows, it's correct.
            assert nb_periods_in_episode >= 1

            #print(f"nb_periods_in_episode={nb_periods_in_episode}")
            nb_periods_per_episode.append(nb_periods_in_episode)

            if plot:
                plt.bar(x=range(len(episode)), height=episode)

            for period_start, period_end in periods:
                # In a period, it rains on every single hour.

                # Find hyétogrammes élémentaires (we call them showers)
                rain = episode[period_start:period_end]

                # Find local minima
                b = np.sign(rain[1:] - rain[0:len(rain)-1])

                # The ndx filtering is here to compress successive
                # "equal pp" hours into only one hour.
                ndx = np.where(b != 0)[0]

                if len(ndx) > 0:
                    # There's a peak
                    # For the flip() see numpy's doc
                    c = np.convolve(
                        b[ndx],
                        np.flip(np.array([-1,1])), mode='valid')
                    # Now we undo the compression done above with ndx[...]
                    # and of course find local minimum with np.where(c==2)
                    crd = ndx[np.where(c == 2)]
                else:
                    # all elements are equal, sot there's no peak.
                    crd = None

                if crd is None or len(crd) == 0:
                    nb_showers_in_period = 1
                else:
                    nb_showers_in_period = len(crd) + 1

                nb_showers_per_period.append(nb_showers_in_period)

                if plot:
                    plt.vlines(x=[period_start-0.5,period_end-1+0.5], ymin=0, ymax=1,color="maroon")
                    if crd is not None:
                        plt.vlines(x=crd+period_start+0.5, ymin=0, ymax=5,color="red")

                if crd is None:
                    indices = [0, len(rain)]
                else:
                    indices = [0] + (crd+1).tolist() + [len(rain)]

                # print(f"elementary_hyetogram based on {indices}")
                # print(rain)
                for i in range(len(indices)-1):
                    elementary_hyetogram = rain[indices[i] : indices[i+1]]
                    #print(f"{indices[i]} {indices[i+1]} {elementary_hyetogram}")

                    shower_duration = len(elementary_hyetogram)
                    shower_mean_intensity = np.sum(elementary_hyetogram) / shower_duration
                    shower_max_to_mean = np.max(elementary_hyetogram) / shower_mean_intensity
                    show_max_pos = np.argmax(elementary_hyetogram) / shower_duration

                    shower_durations.append(shower_duration)
                    shower_mean_intensities.append(shower_mean_intensity)
                    shower_max_to_means.append(shower_max_to_mean)
                    shower_max_to_positions.append(show_max_pos)

            previous_end = None
            for period_start, period_end in periods:
                if previous_end is not None:
                    dry_zone = period_start - previous_end
                    dry_zones.append(dry_zone)
                previous_end = period_end


            #print(f"Dry zones:{dry_zones}")

            if plot:
                plt.xlabel("Time (h)")
                plt.ylabel("Precipitation (mm)")
                plt.title(f"{nb_periods_in_episode} périodes dans cet épisode")
                plt.show()


        if to_file:
            def pa(a):
                return ",".join([str(x) for x in a ]) + "\n"

            fname = f"shypre_{daily_minimum}_{storm_minimum}.txt"
            with open(fname,"w") as fout:
                fout.write(f"{len(rain_episodes)}\n")
                fout.write(pa(nb_periods_per_episode))
                fout.write(pa(nb_showers_per_period))
                fout.write(pa(shower_durations))
                fout.write(pa(shower_mean_intensities))
                fout.write(pa(shower_max_to_positions))
                fout.write(pa(shower_max_to_means))
                fout.write(pa(dry_zones))
                fout.write(pa(episodes_per_year.values()))
            print(f"Wrote {fname}")

            # df = []
            # for k, maxs in shower_max_of_averages.items():
            #     for m in maxs:
            #         df.append( (k+1,m) )
            # pd.DataFrame.from_records(df).to_csv("max_of_averages.csv", index=False)

        if all_periods is not None and len(all_periods) > 0:

            mms = np.array([np.sum(data[b:e]) for b,e in all_periods])
            durations = all_periods[:,1] - all_periods[:,0]

            df_all_periods = pd.concat([
                df_all_periods,
                pd.DataFrame({"StationId":[sid] * all_periods.shape[0],
                              "Date":start_time + all_periods[:,0].astype('timedelta64[h]'),
                              "Duration": durations,
                              "mm":mms,
                              "mean_mm":mms/durations
                              })
            ])


    return df_all_periods


def compute_storms(data, noise_level=0.1, min_precipitations_per_storm=2, include_sids=False, include_mean=False, station_metadata=None):
    # Returns two data frames:
    # - one with the storms
    # - one with the inter arrival time between storms.

    precipitations = []
    precipitations_average = []
    durations = []
    sids = []
    dates = []
    inter_storm_durations = []
    inter_storm_dates = []
    inter_storm_sids = []

    #print(f"{len(data['StationId'].unique())} different stations")

    # We compute station by station because recording
    # timelines are at the station level.
    for sid in data['StationId'].unique():
        if station_metadata:
            print(station_metadata[sid].name)
        else:
            print(sid)

        station_data = data[ data['StationId'] == sid]
        assert len(station_data) > 0, "No storm to process ?"
        v = np.array(station_data["mm"])
        vd = np.array(station_data["Date"])

        noise = v < noise_level
        v[noise] = 0

        # We rely on the fact that records are done every hours
        # So one index position equals one hour.
        nzeros = non_zero_runs(v)

        print(f"Noise hours {len(v)-np.sum(noise)}/{len(v)}")
        print(f"Storms (with small ones)={len(nzeros)}")

        last_storm_end = last_storm_end_date = None
        for i in range(nzeros.shape[0]):
            # Rain event duration (in hour)
            # Float is better suited to further analysis
            # nzeros[i,0] = *index* of a run of non zero precipitation
            # We rely on the fact that recordings happen on *each* *hour* (both are important here)
            # so an index is expressed in hours.
            h_storm_begin = nzeros[i,0]
            h_storm_end = nzeros[i,1]

            d = float(h_storm_end - h_storm_begin)

            # Total precipitation during a rain event
            period = v[h_storm_begin:h_storm_end]
            p = np.sum(period)

            # keep only significant rain events
            # i.e. those with enough rainfall.
            if p >= min_precipitations_per_storm: # Clean up some bad data,
                durations.append(d)
                precipitations.append(p)
                dates.append(vd[h_storm_begin])
                sids.append(sid)
                precipitations_average.append(np.average(period))

                if last_storm_end is not None:
                    iat = h_storm_begin - last_storm_end
                    inter_storm_durations.append(iat)
                    inter_storm_dates.append(last_storm_end_date)
                    inter_storm_sids.append(sid)

                last_storm_end = h_storm_end

                if h_storm_end <= len(vd)-1:
                    last_storm_end_date = vd[h_storm_end]
                else:
                    last_storm_end_date = vd[h_storm_begin]

    # The roundings are here to fix small computation errors
    # that occur during processing (FIXME : find the origin of these errors)
    # The rounding reproduces the rounding in the actual data.
    d = {
        "Date": dates,
        "Duration": np.round(durations),
        "mm": np.round(precipitations,1), # Quantity of water in a given storm
    }
    if include_sids:
        d["StationId"] = sids
    if include_mean:
        d["mean_mm"] = precipitations_average
    storms = pd.DataFrame(d)

    d = {
        "Date": inter_storm_dates,
        "Duration": np.round(inter_storm_durations)
    }
    if include_sids:
        d["StationId"] = inter_storm_sids
    inter_storms = pd.DataFrame(d)

    return storms, inter_storms


def hours_to_duration(n):
    if n < 1:
        return f"{int(n*60)}min"
    elif n < 24:
        return f"{int(n)}h"
    else:
        return f"{int(n)//24}d"


def timedelta_hours(td):
    return td.days * 24 + ((td.seconds+3599) // 3600)

def storms_to_hyetograph(storms):
    base_date = storms["Date"].min()

    timeline = np.zeros(
        timedelta_hours(
            storms["Date"].max() - storms["Date"].min()))

    for idx, row in storms.iterrows():
        b = timedelta_hours(row["Date"] - base_date)
        timeline[b:b+round(row["Duration"])] = row["mm"] / row["Duration"]

    return pd.Series(
        data=timeline,
        index=np.arange(len(timeline))*np.timedelta64(1, 'h') + base_date.to_numpy()
    )

def compute_simple_IDF(data):
    period = np.max(data["Date"]).year - np.min(data["Date"]).year + 1

    year_range = range(np.min(data["Date"]).year,
                       np.max(data["Date"]).year+1)

    DURATIONS = [1,2,3,6,12] + [d*24 for d in [1,2,3,4,5,7,10,15,20,25,30]]
    idf = np.zeros((period, len(DURATIONS)))

    for duration in tqdm(DURATIONS):
        # max_of_avg_intensities = []
        # for year in year_range:
        #     b, e = human_date(f"1/1/{year}"), human_date(f"1/1/{year+1}")
        #     date_filter = (b <= data["Date"]) & (data["Date"] < e)

        #     assert np.count_nonzero(np.isnan(data[date_filter]["mm"])) == 0

        #     avg_intensities = np.convolve( data[date_filter]["mm"],
        #                                    np.ones(duration)/duration,
        #                                    mode="valid")

        #     max_of_avg_intensities.append(np.max(avg_intensities))
        # Each column is intensities (just like the chart)
        # idf[:, DURATIONS.index(duration)] = np.array(sorted(max_of_avg_intensities, reverse=True))

        idf[:, DURATIONS.index(duration)] = sorted(
            np.convolve( data["mm"],
                         np.ones(duration)/duration,
                         mode="valid"), reverse=True)[:period]

    idf = pd.DataFrame(idf, index=range(period,0,-1), columns=DURATIONS)
    idf.rename_axis("Durations", axis="columns", inplace=True)
    idf.rename_axis("Return Period", axis="index", inplace=True)

    return idf

IRM_DURATIONS = [1,2,3,6,12] + [d*24 for d in [1,2,3,4,5,7,10,15,20,25,30]]
def compute_irm_idf(station_pp,
                    storm_mode = False,
                    bm_block_size = "365.25D",
                    return_periods = [2,5,10,15,20,25,30,40,50,75,100,200],
                    durations = IRM_DURATIONS,
                    plot_analysis=False):
    # The default parameters values are IRM values.
    from pyextremes import EVA

    #bm_block_size = "365.25D"
    #bm_block_size = "182.625D"

    idf = np.zeros((len(return_periods), len(durations)))
    models_for_duration = dict() # optimisation cache

    for return_period in return_periods:
        intensities = []
        for duration in durations:

            # We compute average intensities over the duration.
            # However, the e.v. modelling is done on total
            # precipitations in each period. We go back to
            # intensities at the end of the process by
            # dividing by the total duration.

            if duration not in models_for_duration:

                # see IRM article 2011WR011707.pdf
                total_mm_for_d = np.convolve(
                    station_pp.to_numpy(),
                    # The total mm in the duration-h period. The
                    # longer the period, the bigger this is.
                    np.ones(duration), # A one dimensional array of ones. Will sum.
                    mode="valid")

                eva = EVA(
                    pd.Series(
                        total_mm_for_d,
                        index=station_pp.index.to_numpy()[:len(total_mm_for_d)]))
                eva.get_extremes(method="BM", block_size=bm_block_size, extremes_type="high")

                # This is IRM's way, but it doesn't work all the time
                # When it fails, the projections get completely crazy.
                # (like 100x times what they should be)
                eva.fit_model(distribution="genextreme")
                models_for_duration[duration] = eva

                if plot_analysis and duration < 200:
                    print("#########################################")
                    print(f"Duration = {duration}")
                    print("#########################################")
                    print(eva.model)
                    print(eva.model.distribution)
                    eva.plot_diagnostic(alpha=0.95)
                    plt.show()

                    # Show that the GEV fit understimates maximums...
                    # print(f"Duration : {duration}")
                    # print(eva.extremes)
                    # eva.plot_extremes()
                    # plt.show()
                    # plt.close()

                    # x = np.linspace(1e-10,eva.extremes.max()*1.5,1000)
                    # plt.plot(x, eva.model.cdf(x))
                    # #plt.plot(x, eva.model.ppf(eva.model.cdf(x)))
                    # x = 1 - 1/np.array(return_periods)
                    # plt.scatter(eva.model.ppf(x), x, c="red")

                    # for rp in [2,5,10,15,20,25,50,100,200]:
                    #     p = 1 - 1/rp
                    #     plt.text(eva.model.ppf(p), p-0.02, str(rp))

                    # plt.scatter(eva.extremes, eva.model.cdf(eva.extremes))

                    # plt.plot([eva.extremes.max(),eva.extremes.max()],
                    #          [0, eva.model.cdf(eva.extremes.max())], c="black")
                    # plt.text(eva.extremes.max(), 0, "Actual")
                    # plt.plot([eva.model.ppf(1-1/20), eva.model.ppf(1-1/20)],
                    #          [0, 1-1/20], c="red")
                    # plt.text(eva.model.ppf(1-1/20), 0, "Pred")
                    # plt.title(f"Duration {duration}h and values for return period of 20 years")
                    # plt.xlabel("mm/h")
                    # plt.ylabel("Probabilities")
                    # plt.show()
                    # plt.close()

            else:
                # Get cached model.
                eva = models_for_duration[duration]


            # ppf is CDF inverse ! see IRM slides :
            # GEV_d(q) = P(x≤q) = 1-1/T → q(d,T) = GEV_d_inv( 1-1/T )

            # q is a quantile. Wikipedia: quantiles are cut points
            # dividing the range of a probability distribution into
            # continuous intervals with equal probabilities, or
            # dividing the observations in a sample in the same way.
            # GEV_d(q) is the probability of NOT exceeding
            # that quantile. The quantile has been measured as
            # a yearly maximum, so x is a yearly maximum too.
            # So it's the probability for a yearly maximum
            # x to be under q.

            # Now if we look at a return period of T years,
            # the the probability of *exceeding* a quantile (a years' maximum)
            # once in 20 years is 1/20. So the probability
            # of not exceeding it is 1 - 1/20. And the
            # probability of not exceeding it is given by P(x≤q)
            # So we seek q such that P(x≤q) = 1 - 1/T
            # That's because P(x>q) occurs with a probability of 1/T.
            # With EV modelling, we can build GEV_d(q) such
            # that P(x≤q) = GEV_d(q)
            # So GEV_d(q) = P(x≤q) = 1-1/T
            # <=> GEV_d(q) = 1-1/T
            # <=> q = GEV_d_inv( 1-1/T )

            intensity = eva.model.ppf( 1 - 1/return_period) / duration

            intensities.append(intensity)
        idf[return_periods.index(return_period),:] = intensities

    idf2 = pd.DataFrame(idf, index=return_periods, columns=durations)
    idf2.rename_axis("Durations", axis="columns", inplace=True)
    idf2.rename_axis("Return Period", axis="index", inplace=True)


    shapes = dict()
    for duration,eva in models_for_duration.items():
        all_params = eva.distribution.mle_parameters | eva.distribution._fixed_parameters
        print(duration)
        print(all_params)
        shapes[duration] = all_params['c']


    if plot_analysis:
        print(idf2)

    return idf2, shapes


def plot_idf(idf, axs, skip_durations = 0, return_periods=None,
             rp_to_years_factor=1, plot_pp=False):
    # idf: a pandas dataframe. x-axis = durations, y axis = return period, cell = intensities
    # plot_pp : draw precipitations instead of intensities

    if return_periods is None:
        return_periods = idf.index.values

    for ndx, return_period in enumerate(return_periods):
        # Pick a column in the IDF table (to draw a line on the chart...)
        intensities = idf.loc[[return_period]].to_numpy().flatten()[skip_durations:]
        if plot_pp:
            intensities = intensities * idf.columns[skip_durations:]
        axs.plot(range(len(intensities)),
                 intensities,
                 c=cm.viridis(ndx/len(return_periods)),
                 label=f"{return_period*rp_to_years_factor} years")

    xtk = [hours_to_duration(n) for n in idf.columns[skip_durations:]]
    axs.set_xticks(range(len(xtk)), xtk)
    axs.set_xlabel("Durations")
    if not plot_pp:
        axs.set_ylabel("Maximum Intensities (mm/h)")
    else:
        axs.set_ylabel("Maximum Precipitation (mm)")
    axs.legend()


def compute_idf_curves(storms, return_periods):
    period = np.max(storms["Date"]).year - np.min(storms["Date"]).year + 1

    # print(f"Period = {period} years, {len(storms)} storms")

    all_durations = storms["Duration"]

    # Build the IDF curves, one duration at a time

    DURATIONS = np.arange(np.min(all_durations), np.max(all_durations)+1).astype(int)
    idf = np.zeros((period, len(DURATIONS)))

    year_range = range(np.min(storms["Date"]).year,
                       np.max(storms["Date"]).year+1)

    for duration in DURATIONS:
        # We build the maximum intensities on each YEAR
        # So it means we track yearly maximums.

        duration_filter = (storms["Duration"] >= duration) & (storms["Duration"] < duration+1)

        imax = []
        for year in year_range:
            b, e = human_date(f"1/1/{year}"), human_date(f"1/1/{year+1}")
            date_filter = (b <= storms["Date"]) & (storms["Date"] < e)

            # All the storms of the desired duration in the current year.
            a_year_of_storms = storms[ duration_filter & date_filter]

            # There might be none of them ('cos our duration+year filter is rather narrow)...
            if len(a_year_of_storms) > 0:
                # don't forget we look for intensities
                imax.append(np.max(a_year_of_storms['mm']/a_year_of_storms['Duration']))


        # print(f"Duration {duration} for years -> ", ",".join([f"{x:.1f}" for x in imax]))

        # At this point we have the maximum for each year for a GIVEN duration.

        # Smallest intensities first
        imax = np.sort(np.array(imax))

        assert len(imax) <= period
        if len(imax) < period:
            # Now, it is quite possible that for some years we don't have any data.
            # So the imax array may be a bit shorter than expected.
            # In that case, we complete with nan's because we actually have
            # no data to report.

            imax = np.hstack([np.empty(period - len(imax))*np.nan, imax])

        # print(f"{duration} -> ", imax)

        # ranked from lowest intensities (more likely, smaller return period) to highest.

        idf[:, duration-1] = imax

    # idf is an array
    # each column contains the maximum intensities for a given duration, smallest to biggest

    # The all-years maximum happened once in the _period_ interval.
    # So its return period is _period_
    # The next maximum was attained or overflown 2 times in the period
    # So its return period is _period_ - 1 year.
    # etc.

    # Converting to intensities and putting into data frame.

    df = pd.DataFrame(columns=DURATIONS)

    # keep only the return periods
    idf = idf[np.array(return_periods) - 1, :]

    # turn precipitations into intensities
    # idf = idf / np.array(DURATIONS, ndmin=2)

    # numpy to pandas
    df = df.append(pd.DataFrame(idf, columns=df.columns), ignore_index=True)
    df["RetPrd"] = return_periods
    df = df.set_index(["RetPrd"])
    return df




def make_idf_curves(storms):
    print("Deprecated, use compute_idf_curves")

    period = np.max(storms["Date"]).year - np.min(storms["Date"]).year + 1
    RETURN_PERIODS = list(range(1,period+1))

    print(f"Period = {period} years, {len(storms)} storms")
    all_intensities = storms["mm"]
    #all_intensities = storms["mm"] / storms["Duration"]
    all_durations = storms["Duration"]

    # Build the IDF curves, one duration at a time

    DURATIONS = list(range(1,int(np.max(all_durations))+1))
    idf = np.zeros((len(DURATIONS), period))

    year_range = range(np.min(storms["Date"]).year,
                       np.max(storms["Date"]).year+1)

    for duration in tqdm(DURATIONS):
        # We build the maximum intensities on each YEAR

        duration_filter = (storms["Duration"] == duration)

        imax = []
        for year in year_range:
            b, e = datetime(year,1,1), datetime(year+1,1,1)
            date_filter = (b <= storms["Date"]) & (storms["Date"] < e)
            a_year_of_storms = storms[ duration_filter & date_filter]

            if len(a_year_of_storms) > 0:
                # If we have no recrod, then it rained zilch.
                assert( np.sum(a_year_of_storms['mm'].isna()) == 0)
                imax.append(np.max(a_year_of_storms['mm']))

        #print(f"{duration} -> ", imax)

        imax = np.sort(np.array(imax))

        # # Pick the maximum intensities for the given duration; we take 'period' out of them
        # # The period is in years.
        # imax = np.sort(all_intensities[all_durations == duration])

        if len(imax) < period:
            imax = np.hstack([np.zeros(period - len(imax)), imax])
        else:
            imax = imax[len(imax)-period:len(imax)]

        #print(f"{duration} -> ", imax)

        # ranked from lowest intensities (more likely, smaller return period) to highest.

        idf[duration-1, :] = imax
    return pd.DataFrame(idf,
                        columns=[f"{y}Y" for y in RETURN_PERIODS],
                        index=[f"{d}h" for d in DURATIONS])


def draw_idf(ax, idf_table: pd.DataFrame):
    # Plot IDF curve generated by function make_idf_curves(...)
    for i in range(idf_table.shape[1]):
        ax.plot(idf_table.iloc[:,i] / np.array(range(1,idf_table.shape[0]+1)),
                label=f"{list(idf_table)[i]} years",
                c=cm.viridis(i/idf_table.shape[1]))
    ax.set_xlabel("Duration")
    ax.set_ylabel("Intensity (mm/h)")
    #ax.legend()


def simulate(size, mini, frozen_dist, dec, debug=False):
    # Simulate 'size' values, taken randomly
    # out of the frozen distribution 'frozen_dist'.
    #
    # During the simulation, values outside [mini, maxi] (INCLUSIVE)
    # are thrown away. FIXME Possible bias !!!
    #
    # All values are rounded to the dec-th decimal. If dec is None
    # no rounding occur. FIXME Possible bias !!!
    #
    # To compare the simulation data, I build the simulation
    # data exactly as it was selected in the original data.
    #
    # First I generate some samples (more than necessary, but they
    # will be filtered afterwards if not in [mini,maxi])

    rnd = frozen_dist.rvs(size=int(1.5*size))

    # If the number of decimals is specified,
    # then we round the data.

    if dec is not None:
        eps = 10**(-dec) * 10**(-5)
    else:
        eps = 0


    # Clip the data
    rnd = rnd[ mini <= rnd ]

    if debug:
        print(f"Mini before rounding= {mini}, after rounding = {np.min(rnd)}")

    # Round the data to reproduce real data's rounding
    # Pay attention to the left limit to avoid counting
    # half a bin.

    if dec is not None:
        rnd = np.round(rnd - (10**(-dec))/2, decimals=dec)

    # Make sure I have the requested number of samples
    rnd = rnd[:size]

    return rnd

def simulate_like(dataset, frozen_dist, dec=None, debug=False):
    # Simulate data matching a data set, with 'dec' decimals for each
    # sample.

    if dec is None:
        u = np.unique(dataset)
        w = np.min(np.abs(u[:-1] - u[1:]))
        dec = int(-log(w)/log(10))
        print(f"Decimals automatically set to {dec}")

    return simulate( len(dataset), np.min(dataset), frozen_dist, dec, debug)

#s = simulate(1000, 1, 5, stats.lognorm(s=0.5,loc=3), 1)
#print(np.min(s), np.max(s), np.unique(s))

def test_compute_shypre_storms():
    R = [88,0.1,0,10,10,0,0,0,11,11,11,0,0,0.1,0.1,0,1,1,1,1,0,0,0,0]
    R2 = [88,0.1,0,10,10,0,0,0,11,11,11,0,0,0.1,0.1,0,1,2,3,1,2,3,4,0]
    # A day without pp >= 20mm, but 4mm is attained
    SR = [1,0.1,0,1,1,0,0,0,1,1,1,0,0,0.1,0.1,0,1,1,1,1,0,0,0,0]
    # Some noise and some light rian, but not enough to reach the 4mm threshold
    D = [0.1] * 12 + [0.11] * 12
    assert len(R) == len(SR) == len(D) == 24

    def make_df(mm):
        hours = []
        for i in range(len(mm)):
            hours.append(datetime(2022,1, day=1+i//24, hour=i%24))
        sid = [BASTOGNE] * len(mm)

        return pd.DataFrame({
            'StationId': sid,
            'Date': hours,
            'mm': mm})

    storms = compute_shypre_storms(make_df(R + R + R))
    print(storms)
    assert len(storms) == 12

    storms = compute_shypre_storms(make_df(SR + SR + D + SR))
    print(storms)
    assert len(storms) == 0

    storms = compute_shypre_storms(make_df(SR + D + R + D))
    print(storms)
    assert len(storms) == 4

    storms = compute_shypre_storms(make_df(SR + R2 + R2 + D))
    print(storms)
    #assert len(storms) == 4

def test_compute_storms():
    # Test compute storm

    mm = [88, # first storm edge case
          0.1, # inter storm
          13,12,11,
          0,0,0,
          11,11,11,
          0,0,0.1,0.1,
          5] # last storm edge case

    hours = []
    for i in range(len(mm)):
        hours.append(datetime(2022,1,1,hour=2+i))

    sid = [BASTOGNE] * len(mm)

    pp = pd.DataFrame({
        'StationId': sid,
        'Date': hours,
        'mm': mm})

    print(pp)
    print("----------------")
    storms, inter_storms = compute_storms(pp)
    print(storms)
    print("Inter storms:")
    print(inter_storms)

    assert np.all(storms['Duration'] == np.array([1.0, 3, 3, 1]))
    assert np.all(inter_storms['Duration'] == np.array([1.0,3,4]))
    assert np.all( storms['mm'].to_numpy() == np.array([88,36,33,5]))

if __name__=="__main__":
    test_compute_storms()
    exit()

    test_compute_shypre_storms()

    import platform
    if platform.system() == "Windows":
        DATAPATH=r"c:\Users\Administrator\Dropbox\uliege\tfe"
    else:
        DATAPATH="/mnt/data2/uliege/tfe"

    STATIONS = load_station_data(f"{DATAPATH}/stations.csv")
    DATA = load_data(DATAPATH, extended=True)

    broad_select_bang(DATA, STATIONS)

    nb_nans = np.count_nonzero(np.isnan(DATA['mm']))
    print(f"{nb_nans} NaNs left in the data")

    storms, inter_storms = compute_storms(DATA)

    nb_nans = np.count_nonzero(np.isnan(storms['mm']))
    print(f"{nb_nans} NaNs left in the storm data")
